/*
 * file Folder.java
 *
 * Licensed Materials - Property of IBM
 * Restricted Materials of IBM
 *
 * (c) Copyright IBM Corporation 2004, 2010.  All Rights Reserved. 
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or  
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp. 
 */
package javax.wvcm;

import java.util.Map;

import javax.wvcm.PropertyNameList.PropertyName;
import javax.wvcm.ResourceList.ResponseIterator;
import javax.wvcm.WvcmException.ReasonCode;

/**
 * A proxy for a folder resource.
 * 
 * A folder resource is a resource that contains a set of named mappings
 * to other resources, called the "bound members" of that folder.
 * The "members" of a folder are the folder itself and all members
 * of the bound members of the folder. 
 * The semantics of a folder is similar to those of
 * a Windows folder, a Unix directory, or a WebDAV collection.
 * 
 * @since 1.0
 */
public interface Folder extends Resource {

    /**
     * Get the contents of a folder.
     * <p>
     * Postconditions:
     * <li>(read-bound-members): The result contains a proxy for each
     *  bound member of the folder.  The Location of the bound member
     *  whose binding name is "foo" must be
     *  <code>Folder.location().child("foo")</code>.
     * <li>(read-all-members): If deep is true, the result contains a proxy for
     *  every member of the folder identified by this Folder.
     * @param deep whether to retrieve properties for
     * all members of this Folder. 
     * @param feedback the properties to be retrieved.
     *  
     * @return an iterator of {@link Resource} objects that contain properties of
     * the resources that are bound members of this Folder.
     * If <code>deep</code> is <code>true</code>, then
     * all members of the folder (including this Folder) are included.
     * @throws WvcmException
     */
    public ResponseIterator<Resource> doReadMemberList(
            boolean deep,
            Feedback feedback) throws WvcmException;

    /**
     * Removes the binding named bindingName from the folder identified by this Folder.
     * The removal of the binding only guarantees that the resource
     * is no longer accessible via this binding; it does
     * not affect its accessibility via other bindings.
     * If a folder is unbound, no resource is accessible at any
     * location that has the location of the unbound folder as its prefix.
     * <p>
     * Postconditions:
     * <li>(resource-unbound): There is no resource at the location identified by this bindingName.
     * <li>(unbind-activity-reference): If an activity is unbound and has no more bindings to it, 
     *  any reference to that activity
     *  in an {@link ControllableResource#ACTIVITY}, {@link Version#ACTIVITY}, or 
     *  {@link Workspace#CURRENT_ACTIVITY} MUST be removed.
     * <li>(update-predecessor-list): If a version was unbound and has no more bindings to it, the 
     *  server MUST have replaced
     *  any reference to that version in a {@link ControllableResource#PREDECESSOR_LIST} or 
     *  {@link Version#PREDECESSOR_LIST} by a copy of the {@link Version#PREDECESSOR_LIST} of 
     *  the unbound version.
     * <li>(version-history-has-root): If the request unbound the root version of a version 
     *  history and that version has no more bindings to it,
     *  the request MUST have updated the {@link VersionHistory#ROOT_VERSION} of the version history to refer to
     *  another version that is an ancestor of all other remaining versions in that version history.
     *  A result of this postcondition is that every version history will have at least one version,
     *  and the only way to delete all versions is to unbind the version history resource.
     * <li>(delete-version-reference): If a version is unbound and has no more bindings to it, 
     *  any reference to that version in a {@link ControllableResource#MERGE_LIST} or 
     *  {@link ControllableResource#AUTO_MERGE_LIST} property MUST be removed.
     * <li>(delete-version-set): If the request unbound a version history and has no more 
     *  bindings to it, the request MUST have unbound all versions in the VersionList of 
     *  that version history, and MUST have satisfied the postconditions for version deletion.
     * <li>(unbind-workspace-reference): If a workspace is unbound and has no more bindings
     *  to it, any reference to that workspace in an activity's {@link Activity#CURRENT_WORKSPACE_LIST}
     *  MUST be removed.
     * <li>(auto-checkout-parent): If the resource identified by <code>bindingName</code>
     *  is version-controlled, and if this folder was a checked-in version-controlled folder
     *  the request MAY have automatically checked out this folder.
     * 
     * @param feedback Specifies optional feedback to the caller.
     * @return a new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#NOT_FOUND}:
     * There must be a binding named <code>bindingName</code> in this Folder.
     * <li>{@link ReasonCode#METHOD_NOT_SUPPORTED}:
     *  A server MAY fail an attempt to unbind a version history or a version.
     * <li>Since this method may cause a checkout to be attempted on this resource,
     *  see {@link ControllableResource#doCheckin} for additional reason codes.
     */
    public Folder doUnbindChild(String bindingName, Feedback feedback) throws WvcmException;

    /** Boolean flags for the doRebind method */
    public static enum RebindFlag 
    {
        /** 
         * Indicates whether to overwrite an existing binding in the destination folder
         */
        OVERWRITE("overwrite"); //$NON-NLS-1$

        private RebindFlag(String flagName) { _flagName = flagName; }

        /**
         * Returns a string representation of this flag suitable for diagnostics.
         */
        @Override
        public String toString() { return _flagName; }

        /**
         * The string form of this flag.
         */
        private final String _flagName;

    }


    /**
     * Removes the binding named <code>sourceBindingName</code> from a resource in the folder
     * identified by <code>sourceFolder</code> and adds the binding named <code>bindingName</code>
     * to that resource in the folder identified by this Folder.
     * The content and properties of the rebound resource are not modified
     * by this request, except for the properties that are location dependent.
     * <p>
     * Postconditions:
     * <li>(removed-from-source-parent): The binding named <code>sourceBindingName</code>
     *  is removed from the folder identified by <code>sourceFolder</code>.
     *  If this was the only binding to this resource in the source folder,
     *  the {@link Resource#PARENT_LIST} of this resource
     *  no longer contains a reference to the source folder.
     * <li>(added-to-destination-parent): The folder identified by the this Folder
     *  is modified to have a binding named <code>bindingName</code> to the rebound resource.
     *  The {@link Resource#PARENT_LIST} of the rebound resource
     *  contains a reference to the folder identified by this Folder.
     * <li>(preserve-properties): The property values of the rebound resource
     *  MUST NOT have been modified by the request unless this specification states otherwise.
     * <li>(workspace-member-moved): If the rebound resource did not identify a workspace,
     *  the {@link ControllableResource#WORKSPACE} property of the rebound resource
     *  MUST have been updated to have the same value as 
     *  the Workspace of this Folder.
     * <li>(workspace-moved): If the rebound resource was a workspace,
     *  any reference to that workspace in a {@link ControllableResource#WORKSPACE} property MUST have been updated
     *  to refer to the new location of that workspace.
     * <li>(update-checked-out-reference): If a checked-out resource is rebound, any reference
     *  to that resource in an {@link Activity#ACTIVITY_CHECKOUT_LIST} property MUST be updated
     *  to refer to the new location of that resource.
     * <li>(update-workspace-reference): If the rebound resource was a workspace,
     *  any reference to that workspace in a CurrentWorkspaceList property
     *  MUST be updated to refer to the new location of that workspace.
     * <li>(update-activity-reference): If the rebound resource was an activity, 
     *  any reference to that activity in an {@link ControllableResource#ACTIVITY},
     *  {@link Version#ACTIVITY}, or {@link Workspace#CURRENT_ACTIVITY} 
     *  MUST be updated to refer to the new location of that activity.
     *  <li>(auto-checkout-destination-parent): If the resource identified by 
     *  <code>sourceBindingName</code> in <code>sourceFolder</code> is a version-controlled
     *  resource, and if this folder was a checked-in folder
     *  the request MUST have automatically checked out this folder.
     *  <li>(auto-checkout-source-parent): If the resource identified by 
     *  <code>sourceBindingName</code> in <code>sourceFolder</code> is a version-controlled
     *  resource, and if the <code>sourceFolder</code> was a checked-in folder,
     *  the request MUST have automatically checked out the source folder.
     *  
     * @param bindingName the new binding name of the resource being rebound.
     * @param sourceFolder the previous parent folder of the rebound resource.
     * @param sourceBindingName the original binding name of the rebound resource in the source folder.
     * @param flags Array of boolean flags (may be null):
     * <li>{@link RebindFlag#OVERWRITE}:
     *  Indicates whether an existing binding in this folder
     *  will be overwritten rather than the request failing.
     * @param feedback Specifies optional feedback to the caller.
     * @return a new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#BAD_ARGUMENT_TYPE}:
     * The resource identified by the <code>sourceFolder</code> argument must be a folder.
     * <li>{@link ReasonCode#CANNOT_OVERWRITE}:
     *  If there already is a binding named <code>bindingName</code> in this Folder,
     *  and <code>overwrite</code> is <code>false</code>, the request MUST fail.
     * <li>{@link ReasonCode#METHOD_NOT_SUPPORTED}:
     *  The folder does not support rebinding children.
     * <li>Since this method may cause a checkout to be attempted on this resource,
     *  see {@link ControllableResource#doCheckin} for additional reason codes.
     */
    public Folder doRebindChild(String bindingName,
            Folder sourceFolder, String sourceBindingName,
            RebindFlag[] flags, Feedback feedback)
    throws WvcmException;

    /**
     * Removes all binding to the resource
     * identified by <code>source</code> and adds the binding named <code>bindingName</code>
     * to that resource in the folder identified by this Folder.
     * The content and properties of the rebound resource are not modified
     * by this request, except for the properties that are location dependent.
     * @see #doRebindChild
     */
    public Folder doRebindAll(String bindingName,
            Resource source,
            RebindFlag[] flags, Feedback feedback)
    throws WvcmException;

    /** Boolean flags for the doBind method */
    public static enum BindFlag 
    {
        /** 
         * Indicates whether to overwrite an existing binding in the destination folder
         */
        OVERWRITE("overwrite"), //$NON-NLS-1$

        /** 
         * Indicates whether to automatically unbind the resource from its original
         * folder if multiple bindings to the resource is not supported.
         */
        AUTO_UNBIND("auto-unbind"); //$NON-NLS-1$

        private BindFlag(String flagName) { _flagName = flagName; }

        /**
         * Returns a string representation of this flag suitable for diagnostics.
         */
        @Override
        public String toString() { return _flagName; }

        /**
         * The string form of this flag.
         */
        private final String _flagName;

    }


    /**
     * Creates a binding named <code>bindingName</code> in the folder identified by this Folder
     * to the resource identified by <code>resource</code>.
     * The content and location-independent properties of the resource are not modified by this request.
     * <p>
     * Postconditions:
     * <li>(preserve-properties): The property values of the resource identified by this Resource
     *  MUST NOT have been modified by this request unless this specification states otherwise.
     * <li>(new-binding): There MUST be a binding named <code>bindingName</code> in the folder
     * identified by this Folder to the resource identified by <code>resource</code>.
     * <li>(auto-checkout-parent): If <code>resource</code> identifies
     *  a controlled resource, and if this Folder identifies a checked-in controlled folder,
     *  this folder MAY have been checked-out by this request.
     * 
     * @param bindingName the name of the new binding in this Folder to the resource
     * identified by <code>resource</code>.
     * @param resource the resource being bound.
     * @param flags Array of boolean flags (may be null):
     * <li>{@link BindFlag#OVERWRITE}:
     *  Indicates whether an existing binding in this folder
     *  will be overwritten rather than the request failing.
     * <li>{@link BindFlag#AUTO_UNBIND}:
     *  Indicates whether the resource will be unbound from its original folder
     *  rather than the request failing because multiple bindings to the same resource are not supported.
     * @param feedback Specifies optional feedback to the caller.
     * @return a new proxy for this resource, whose properties are specified by feedback.
     * @throws WvcmException ReasonCode:
     * <li>{@link ReasonCode#NOT_FOUND}:
     * There must be a binding named <code>bindingName</code> in this Folder,
     * and the folder identified by <code>sourceFolder</code> must exist.
     * <li>{@link ReasonCode#METHOD_NOT_SUPPORTED}:
     *  If the resource identified by <code>resource</code> does not
     *  support multiple bindings to it, and {@link BindFlag#AUTO_UNBIND} is not
     *  specified, the request MUST fail.
     * <li>{@link ReasonCode#NO_CROSS_SERVER_BINDING}:
     *  If the resource identified by <code>resource</code> is on another server from
     *  the folder identified by this Folder, and the server for this Folder
     *  does not support cross-server bindings, the request MUST fail.
     * <li>{@link ReasonCode#CANNOT_OVERWRITE}:
     *  If there already is a binding named <bindingName> in the folder identified
     *  by this Folder, and <code>overwrite</code> is <code>false</code>, the request MUST fail.
     * <li>{@link ReasonCode#CYCLE_NOT_ALLOWED}:
     *  If the server does not support cycles in the location namespace,
     *  and if <code>resource</code> identifies a folder,
     *  and if the folder identified by this Folder is a member of that folder,
     *  then the request MUST fail.
     * <li>Since this method may cause a checkout to be attempted on this resource,
     *  see {@link ControllableResource#doCheckin} for additional reason codes.
     */
    public Folder doBindChild(
            String bindingName,
            Resource resource,
            BindFlag[] flags, 
            Feedback feedback)
    throws WvcmException;

    /**
     * The map of (name, resource) pairs identifying the resources bound to
     * this folder as immediate children and the names by which they are known
     * in this folder. A map entry's key is the binding name (String) and its value
     * is the bound member (Resource).
     * Unlike the locations returned by
     * {@link #doReadMemberList}, the locations of the child resources in the
     * {@link #CHILD_MAP} <i>do not</i> need to be simple extensions of the
     * location of this folder. Thus, if <code>segment</code> is a
     * binding name key from the {@link #CHILD_MAP} property of
     * Folder <code>foo</code>, and <code>member</code> is a Resource value for
     * that key:
     * <ul>
     * <li><code>segment</code> <i>does not</i> need to
     * equal <code>member.location().lastSegment()</code>
     * <li><code>member.location().getParent()</code>
     * <i>does not</i> need to equal <code>foo.location()</code>
     * <li>although <code>foo.location().child(binding)</code>
     * <i>must be</i> a valid location for the bound resource, it <i>does not</i>
     * need to equal <code>member.location()</code>
     * </ul>
     * Since the locations are less constrained, computation
     * of {@link #CHILD_MAP} may be more efficient than {@link #doReadMemberList}.
     */
    final PropertyName<Map<String, Resource>> CHILD_MAP = 
        new PropertyName<Map<String, Resource>>("child-map"); //$NON-NLS-1$

    /**
     * Get the {@link #CHILD_MAP} property.
     * 
     * @return {@link #CHILD_MAP} property.
     * @throws WvcmException if this Folder was not created with
     * {@link #CHILD_MAP} as a wanted property.
     */
    public Map<String, Resource> getChildMap() throws WvcmException;

    /**
     * The list of resources bound to this folder as immediate children.
     * This list contains the same resources as the values of {@link #CHILD_MAP}.
     * This may be a more efficient way of accessing the list
     * of children when their binding names are not required.
     */
    final PropertyName<ResourceList<Resource>> CHILD_LIST = 
        new PropertyName<ResourceList<Resource>>("child-list"); //$NON-NLS-1$

    /**
     * Get the {@link #CHILD_LIST} property.
     * 
     * @return {@link #CHILD_LIST} property.
     * @throws WvcmException if this Folder was not created with
     * {@link #CHILD_LIST} as a wanted property.
     */
    public ResourceList<Resource> getChildList() throws WvcmException;


}
